# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from maite._internals.testing.project import (
    CompletenessSection,
    ModuleScan,
    ModuleScanResults,
    Symbol,
    SymbolCounts,
    get_public_symbols,
    import_public_symbols,
)

__all__ = [
    "CompletenessSection",
    "ModuleScan",
    "ModuleScanResults",
    "Symbol",
    "SymbolCounts",
    "get_public_symbols",
    "import_public_symbols",
]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# flake8: noqa

from typing import List, Tuple

import hypothesis.strategies as st
import pytest
from hypothesis import given, settings

from maite._internals.testing.pyright import md_to_code, rst_to_code

src1 = """.. tab-set::

   .. tab-item:: dataclass-based config

      .. code-block:: python
         :caption: aa
      
         from dataclasses import dataclass
      
         @dataclass
         class Builds_DNN:
             input_size: int
             output_size: int
             layer_widths: Tuple[int, ...] = (5, 10, 5)
             device: str = "cpu"
             _target_: str = "vision.model.DNN"

   .. tab-item:: yaml-based config

      .. code-block:: yaml
         :caption: blah
      
         _target_: vision.model.DNN
         input_size: ???
         output_size: ???
         layer_widths:
         - 5
         - 10
         - 5
"""
expected1 = """
from dataclasses import dataclass

@dataclass
class Builds_DNN:
    input_size: int
    output_size: int
    layer_widths: Tuple[int, ...] = (5, 10, 5)
    device: str = "cpu"
    _target_: str = "vision.model.DNN"

"""

src2 = """
.. code-block:: pycon

   >>> from pathlib import Path 
   >>> def print_file(x: int):
   ...     with x.open("r") as f: 
   ...         print(f.read())
"""

src3 = """
    .. code-block:: python
       :caption: blah
       :foo: a
       :bar: b

       from pathlib import Path
   
       def print_file(x: Path):
           with x.open("r") as f: 
               print(f.read())
"""

expected2 = """from pathlib import Path 

def print_file(x: int):
    with x.open("r") as f: 
        print(f.read())
"""
expected3 = """
from pathlib import Path

def print_file(x: Path):
    with x.open("r") as f: 
        print(f.read())
"""


@pytest.mark.parametrize(
    "src,expected",
    [
        pytest.param(src1, expected1, id="src1"),
        pytest.param(src2, expected2, id="src2"),
        pytest.param(src2 * 3, "\n".join([expected2] * 3), id="src2 repeat"),
        pytest.param(src3 * 3, "\n".join([expected3] * 3), id="src3 repeat"),
        pytest.param(src1 + src2, "\n".join([expected1, expected2]), id="src1+src2"),
        pytest.param("", "", id="empty input"),
    ],
)
def test_rst_parsing_manual(src: str, expected: str):
    assert strip_interspacing(rst_to_code(src)) == strip_interspacing(expected)


distractors = [
    "",
    "\n",
    "python",
    "pycon",
    ">>> x = 1",
    "word.",
    ".. code-block::",
    "  .. code-block::",
    """.. code-block:: shell

   blah blah""",
    """.. code-block:: shell
   :caption: hello

   blah blah""",
]


def strip_interspacing(x: str):
    return "\n".join((s for s in x.splitlines() if s))


import textwrap


@settings(max_examples=500)
@given(
    blocks=st.lists(
        st.sampled_from([(src1, expected1), (src2, expected2), (src3, expected3)]),
        min_size=1,
        max_size=5,
    ),
    data=...,
    indent_level=st.integers(0, 9),
)
def test_rst_parsing(
    blocks: List[Tuple[str, str]], data: st.DataObject, indent_level: int
):
    """Combines known src blocks (with associated expected outputs), joined by
    different patterns of joining text"""
    src_blocks = [b[0] for b in blocks]
    expected_blocks = [b[1] for b in blocks]

    num_blocks = len(blocks)
    dummy_blocks = data.draw(
        st.lists(
            st.sampled_from(distractors),
            min_size=num_blocks + 1,
            max_size=num_blocks + 1,
        )
    )
    src = "\n".join(a + b for a, b in zip(dummy_blocks, src_blocks + [""]))
    src = textwrap.indent(src, " " * indent_level)
    expected = strip_interspacing("\n".join([s.strip() for s in expected_blocks]))
    out = strip_interspacing(rst_to_code(src))

    assert out == expected


md_src1 = """.. tab-set::

   .. tab-item:: dataclass-based config

      ```python
      
         from dataclasses import dataclass
      
         @dataclass
         class Builds_DNN:
             input_size: int
             output_size: int
             layer_widths: Tuple[int, ...] = (5, 10, 5)
             device: str = "cpu"
             _target_: str = "vision.model.DNN"
      ```

   ```yaml
         _target_: vision.model.DNN
         input_size: ???
         output_size: ???
         layer_widths:
         - 5
         - 10
         - 5
   ```
"""
md_expected1 = """
from dataclasses import dataclass

@dataclass
class Builds_DNN:
    input_size: int
    output_size: int
    layer_widths: Tuple[int, ...] = (5, 10, 5)
    device: str = "cpu"
    _target_: str = "vision.model.DNN"

"""

md_src2 = """
```pycon

   >>> from pathlib import Path 
   >>> def print_file(x: int):
   ...     with x.open("r") as f: 
   ...         print(f.read())
```
"""

md_src3 = """
    ````
    an example
    ```python
    x = 2
    ```
    ````
    ```python

       from pathlib import Path
   
       def print_file(x: Path):
           with x.open("r") as f: 
               print(f.read())
    ```
"""

md_expected2 = """from pathlib import Path 

def print_file(x: int):
    with x.open("r") as f: 
        print(f.read())
"""
md_expected3 = """
from pathlib import Path

def print_file(x: Path):
    with x.open("r") as f: 
        print(f.read())
"""


@pytest.mark.parametrize(
    "src,expected",
    [
        pytest.param(md_src1, md_expected1, id="src1"),
        pytest.param(md_src2, md_expected2, id="src2"),
        pytest.param(md_src2 * 3, "\n".join([md_expected2] * 3), id="src2 repeat"),
        pytest.param(md_src3 * 3, "\n".join([md_expected3] * 3), id="src3 repeat"),
        pytest.param(
            md_src1 + md_src2, "\n".join([md_expected1, md_expected2]), id="src1+src2"
        ),
        pytest.param("", "", id="empty input"),
    ],
)
def test_md_parsing_manual(src: str, expected: str):
    assert strip_interspacing(md_to_code(src)) == strip_interspacing(expected)



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# flake8: noqa
import json
import re
from pathlib import Path

import pytest

from maite.testing.pyright import list_error_messages, pyright_analyze


def test_pyright_catches_errors():
    def f(x: str):
        x + 1
        x.append(2)

    results = pyright_analyze(f)
    assert results[0]["summary"]["errorCount"] == 2


def test_pyright_scans_clean():
    def f():
        def g(x: int):
            ...

        g(1)
        g(2)

    x = pyright_analyze(f)
    assert x[0]["summary"]["errorCount"] == 0


def test_pyright_basic():
    def f(x):
        ...

    results = pyright_analyze(f, type_checking_mode="basic")
    assert results[0]["summary"]["errorCount"] == 0


def test_pyright_strict():
    def f(x):  # type of x is unknown, missing return annotation
        ...

    results = pyright_analyze(f, type_checking_mode="strict")
    assert results[0]["summary"]["errorCount"] > 0


@pytest.mark.usefixtures("cleandir")
def test_python_version():
    code = """def f(x: list[int]):  # using builtin type as generic is new to 3.9
        ...
    """
    with open("f.py", "w") as file:
        file.writelines(code)

    py38 = pyright_analyze("./", python_version="3.8")
    py39 = pyright_analyze("./", python_version="3.9")
    assert py38[0]["summary"]["errorCount"] == 1, py38
    assert py39[0]["summary"]["errorCount"] == 0


def test_scan_path_to_code():
    import maite

    results = pyright_analyze(
        Path(maite.__file__).parent, report_unnecessary_type_ignore_comment=True
    )
    assert len(results[0]["generalDiagnostics"]) >= 0


def test_preamble():
    import math

    def g():
        math.acos(1)

    results = pyright_analyze(g, preamble="import math")
    assert results[0]["summary"]["errorCount"] == 0


def test_scan_docstring_raises_on_path():
    with pytest.raises(
        ValueError, match=re.escape(r"`scan_docstring=True` can only be specified")
    ):
        pyright_analyze(Path.cwd(), scan_docstring=True)


def test_scan_docstring():
    def f():
        """
        Examples
        --------
        >>> x = 1
        >>> y = x + 'a'  # pyright should catch this
        """
        return

    results = pyright_analyze(f, scan_docstring=True)
    assert results[0]["summary"]["errorCount"] == 1
    (message,) = [
        d["message"]
        for d in results[0]["generalDiagnostics"]
        if d["severity"] == "error"
    ]
    assert message.startswith('Operator "+" not supported for types')


rst_good_1 = """
    .. code-block:: python

       from pathlib import Path
   
       def print_file(x: Path) -> None:
           with x.open("r") as f: 
               print(f.read())
"""

rst_good_2 = """
    .. code-block:: pycon

       >>> from pathlib import Path
       >>>
       >>> def print_file(x: Path) -> None:
       ...     with x.open("r") as f: 
       ...         print(f.read())
"""

rst_bad_1 = """
    .. code-block:: python

       from pathlib import Path
   
       def print_file(x: int) -> None:
           with x.open("r") as f: 
               print(f.read())
"""

rst_bad_2 = """
    .. code-block:: pycon

       >>> from pathlib import Path
       >>>
       >>> def print_file(x: int) -> None:
       ...     with x.open("r") as f: 
       ...         print(f.read())
"""


@pytest.mark.usefixtures("cleandir")
@pytest.mark.parametrize(
    "src, expected_num_error",
    [
        (rst_good_1, 0),
        (rst_good_2, 0),
        (rst_bad_1, 1),
        (rst_bad_2, 1),
    ],
)
def test_scan_rst(src: str, expected_num_error: int):
    Path("file.rst").write_text(src)  # file will be written to a tmp dir
    results = pyright_analyze("file.rst")
    assert (
        results[0]["summary"]["errorCount"] == expected_num_error
    ), list_error_messages(results[0])


md_good_1 = """
    blah blah

    ```python
       from pathlib import Path
   
       def print_file(x: Path) -> None:
           with x.open("r") as f: 
               print(f.read())
    ```
    ya ya
    ````
    ```python
    just an example in a literal block
    ```
    ````
    ``python
    not a block
    ``
"""

md_good_2 = """
    ```pycon

    >>> from pathlib import Path
    >>>
    >>> def print_file(x: Path) -> None:
    ...     with x.open("r") as f: 
    ...         print(f.read())
    ```
    ``pycon
    not a block
    ``
"""

md_bad_1 = """
    ```python
       from pathlib import Path
   
       def print_file(x: int) -> None:
           with x.open("r") as f: 
               print(f.read())
    ```
"""

md_bad_2 = """

    ```pycon

    >>> from pathlib import Path
    >>>
    >>> def print_file(x: int) -> None:
    ...     with x.open("r") as f: 
    ...         print(f.read())
    ```
"""


@pytest.mark.usefixtures("cleandir")
@pytest.mark.parametrize(
    "src, expected_num_error",
    [
        (md_good_1, 0),
        (md_good_2, 0),
        (md_bad_1, 1),
        (md_bad_2, 1),
    ],
)
def test_scan_md(src: str, expected_num_error: int):
    Path("file.md").write_text(src)  # file will be written to a tmp dir
    results = pyright_analyze("file.md")
    assert (
        results[0]["summary"]["errorCount"] == expected_num_error
    ), list_error_messages(results[0])


@pytest.mark.filterwarnings("ignore:the imp module is deprecate")
@pytest.mark.filterwarnings("ignore:Jupyter is migrating its paths")
@pytest.mark.parametrize("src, expected_num_error", [("1 + 'a'", 1), ("1 + 2", 0)])
@pytest.mark.usefixtures("cleandir")
def test_scan_ipynb(src, expected_num_error):
    import jupytext

    jupytext.write(jupytext.reads(src, fmt=".py"), "file.ipynb", fmt=".ipynb")

    results = pyright_analyze("file.ipynb")
    assert (
        results[0]["summary"]["errorCount"] == expected_num_error
    ), list_error_messages(results[0])


@pytest.mark.usefixtures("cleandir")
def test_scan_doesnt_clobber_preexisting_pyright_config():
    d = Path.cwd()
    d.mkdir(exist_ok=True)

    config = d / "pyrightconfig.json"
    file_ = d / "file.py"
    pyright_target = file_

    config.write_text(json.dumps({"reportUnnecessaryTypeIgnoreComment": False}))
    expected_config = config.read_text("utf-8")
    file_.write_text("x = 1 + 1  # type: ignore")

    results = pyright_analyze(
        pyright_target,
        pyright_config={"reportUnnecessaryTypeIgnoreComment": True},
    )
    post_run_config = config.read_text("utf-8")

    assert (
        results[0]["summary"]["errorCount"] == 1
        and 'Unnecessary "# type: ignore" comment'
        in results[0]["generalDiagnostics"][0]["message"]
    )

    assert expected_config == post_run_config


@pytest.mark.parametrize("suffix", [".py", ".ipynb", ".rst", "/"])
def test_analyze_missing_file(suffix):
    with pytest.raises(FileNotFoundError, match=r"Cannot be scanned by pyright."):
        pyright_analyze(f"moo{suffix}")


def test_bad_path_to_pyright():
    def f():
        ...

    bad_path = Path("not/a/path/pyright")
    with pytest.raises(
        FileNotFoundError, match=re.escape(f"{str(bad_path)} – doesn't exist.")
    ):
        pyright_analyze(f, path_to_pyright=Path(bad_path))


@pytest.mark.usefixtures("cleandir")
def test_unsupported_file_type():
    f = Path.cwd() / "file.txt"
    f.touch()
    with pytest.raises(ValueError, match=re.escape("File type .txt not supported")):
        pyright_analyze(f)


def test_list_error_messages():
    def f(x: int):
        return x.lower()

    results = pyright_analyze(f)[0]
    listed_errors = list_error_messages(results)
    assert len(listed_errors) == 1
    assert listed_errors[0].startswith(
        '(line start) 1: Cannot access member "lower" for type "int"'
    )



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from __future__ import annotations

import itertools
from typing import List

import pytest

from maite._internals.interop.provider_registry import (
    create_provider,
    get_provider_type,
    list_providers,
    register_provider,
)
from maite._internals.protocols.generic import Dataset, Metric, Model
from maite._internals.protocols.task_aliases import TaskName
from maite.errors import InternalError, InvalidArgument


class AModel:
    def __call__(self, arg: list):
        return arg


class AMetric:
    def init(self):
        self.pred = 0
        self._count = 0

    def reset(self):
        pass

    def update(self, pred, target):
        self.pred = pred
        self.target = target
        self._count += 1

    def compute(self):
        return {"metrics": self.pred / self._count}


class ADataset:
    def __init__(self, data):
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]


class BaseProvider:
    def __init__(self, model_names):
        self._names = model_names

    def help(self, name):
        return name


class FullModelProvider(BaseProvider):
    def list_models(
        self,
        *,
        filter_str: str | List[str] | None = None,
        model_name: str | None = None,
        task: TaskName | None = None,
    ):
        return [m for m in self._names]

    def load_model(self, model_name: str, task: TaskName, **kwargs) -> Model:
        return AModel()


class FullMetricProvider(BaseProvider):
    def list_metrics(self):
        return [m for m in self._names]

    def load_metric(self, metric_name: str) -> Metric:
        return AMetric()


class FullDatasetProvider(BaseProvider):
    def list_datasets(self):
        return [m for m in self._names]

    def load_dataset(
        self,
        *,
        dataset_name: str,
        task: TaskName,
        split: str | None = None,
    ) -> Dataset:
        return ADataset([x for x in range(len(dataset_name))])


class UniqueNameGen:
    __name__ = "UniqueNameGen"

    def __init__(self):
        self.counter = itertools.count(0)

    def __call__(self, prefix):
        return f"{prefix}_{next(self.counter)}"


_name_gen_instance = UniqueNameGen()


@pytest.fixture
def name_gen():
    return _name_gen_instance


def test_register_provider(name_gen):
    # Preamble, capture "native providers"
    native_providers = set(list_providers())
    # Part 1 - test registration

    # no error
    register_provider(FullModelProvider, key=name_gen("provider"))

    # no protocol - Error
    with pytest.raises(InvalidArgument):
        register_provider(BaseProvider, key=name_gen("provider"))  # type: ignore
    # no key - Error
    with pytest.raises(InvalidArgument):
        register_provider(FullModelProvider)

    # define key at class scope
    class ProviderWithKey(FullModelProvider):
        PROVIDER_REGISTRY_KEY = name_gen("provider")

    # Good
    register_provider(ProviderWithKey)

    # Re-register -> Warning, but okay
    with pytest.warns(
        Warning, match=f"under name {ProviderWithKey.PROVIDER_REGISTRY_KEY}"
    ):
        register_provider(ProviderWithKey)

    new_registered_providers = set(list_providers()) - native_providers
    # we performed 3 successful registrations above, but the final one overwrites an existing one,
    # so there should be 2 non-native providers registered
    assert len(new_registered_providers) == 2


def test_list_providers(name_gen):
    # generate 6 names
    names = [name_gen("provider") for _ in range(6)]
    ds_provider = names[0]
    met_provider = names[1]
    model_provider = names[2]
    ds_met_provider = names[3]
    ds_model_provider = names[4]
    any_provider = names[5]

    # register some providers
    register_provider(FullDatasetProvider, key=ds_provider)
    register_provider(FullMetricProvider, key=met_provider)
    register_provider(FullModelProvider, key=model_provider)

    # create some types satisfying multiple provider protocols
    class DSAndMetricProvider(FullDatasetProvider, FullMetricProvider):
        pass

    class DSAndModelProvider(FullDatasetProvider, FullModelProvider):
        pass

    class DSMetricModelProvider(
        FullDatasetProvider, FullMetricProvider, FullModelProvider
    ):
        pass

    register_provider(DSAndMetricProvider, key=ds_met_provider)
    register_provider(DSAndModelProvider, key=ds_model_provider)
    register_provider(DSMetricModelProvider, key=any_provider)

    bare_list = list_providers()
    # all registered names should be in the unfiltered list
    assert all(n in bare_list for n in names)
    # limit to model
    model_prov_list = list_providers(enforce_protocol="model")
    assert all(
        n in model_prov_list for n in (model_provider, ds_model_provider, any_provider)
    )
    assert all(
        n not in model_prov_list for n in (ds_provider, met_provider, ds_met_provider)
    )
    # limit to dataset
    ds_prov_list = list_providers(enforce_protocol="dataset")
    assert all(
        n in ds_prov_list
        for n in (ds_provider, ds_met_provider, ds_model_provider, any_provider)
    )
    assert all(n not in ds_prov_list for n in (met_provider, model_provider))
    # limit to metric
    met_prov_list = list_providers(enforce_protocol="metric")
    assert all(
        n in met_prov_list for n in (met_provider, ds_met_provider, any_provider)
    )
    assert all(
        n not in met_prov_list for n in (ds_provider, model_provider, ds_model_provider)
    )


def test_get_provider_type(name_gen):
    p1_name = name_gen("provider")
    register_provider(FullDatasetProvider, key=p1_name)

    p1_type = get_provider_type(p1_name)
    assert p1_type is FullDatasetProvider

    # raises when it can't match the key
    with pytest.raises(InvalidArgument):
        get_provider_type(name_gen("provider"))

    # raises when the key does not match the protocol provided
    with pytest.raises(InvalidArgument):
        get_provider_type(p1_name, enforce_protocol="model")


def test_create_provider(name_gen):
    # generate and register a provider with kwargs for __init__
    provider_name = name_gen("provider")

    @register_provider(key=provider_name)
    class ProviderWithKwargInit(FullDatasetProvider):
        def __init__(self, names, *, a=None, b=None, c=None):
            super().__init__(names)
            self.a = a
            self.b = b
            self.c = c

    # generate some args/kwargs
    names_arg = [name_gen("name") for _ in range(4)]
    provider_kwargs = {"a": "foo", "b": "bar", "c": "baz"}
    # find *and* instantiate the provider type
    my_provider = create_provider(
        provider_name, provider_args=(names_arg,), provider_kwargs=provider_kwargs
    )
    assert isinstance(my_provider, ProviderWithKwargInit)
    assert my_provider.a == provider_kwargs["a"]
    assert my_provider.b == provider_kwargs["b"]
    assert my_provider.c == provider_kwargs["c"]
    provider_datasets = my_provider.list_datasets()
    assert all(name in provider_datasets for name in names_arg)


def test_registry_instantiate_error():
    from maite._internals.interop.provider_registry import _ProviderRegistry

    with pytest.raises(InternalError):
        _ = _ProviderRegistry()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))
import maite

# -- Project information -----------------------------------------------------

project = "maite"
copyright = "2024 Massachusetts Institute of Technology"
author = "Ryan Soklaski, Justin Goodwin, Michael Yee"

# The short X.Y version
version = ".".join(maite.__version__.split(".")[:-1])
# The full version, including alpha/beta/rc tags
release = maite.__version__


# -- General configuration ---------------------------------------------------

REPO_URL = "https://github.com/mit-ll-ai-technology/maite"

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.autosummary",
    "numpydoc",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_codeautolink",
    "myst_parser",
    "sphinx.ext.doctest",
]

# set autosummary_generate to True to repopulate autosummary-generated rst files
# (needed only after adding/removing objects (e.g. classes/functions) under
# 'autosummary' directives in 'api_reference.rst'.
autosummary_generate = False
numpydoc_show_inherited_class_members = False
numpydoc_class_members_toctree = False
# prevents build warnings resulting from Sphinx not finding generated documentation for
# attributes of NumpyDocResults and PyrightOutput (these attributes have no docstrings
# in the source, and thus an autodoc-populated page is never generated for them.)
# Class members are typically documented directly in protocol class docstrings where
# no member-specific page is needed.

# Strip input prompts:
# https://sphinx-copybutton.readthedocs.io/en/latest/#strip-and-configure-input-prompts-for-code-cells
copybutton_prompt_text = r">>> |\.\.\. |\$ |In \[\d*\]: | {2,5}\.\.\.: | {5,8}: "
copybutton_prompt_is_regexp = True


default_role = "py:obj"

autodoc_typehints = "none"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html
# A dictionary of external sites
#   alias ->  (base-URL, prefix)
extlinks = {
    "commit": (REPO_URL + "commit/%s", "commit %s"),
    "gh-file": (REPO_URL + "blob/master/%s", "%s"),
    "gh-link": (REPO_URL + "%s", "%s"),
    "issue": (REPO_URL + "issues/%s", "issue #%s"),
    "pull": (REPO_URL + "pull/%s", "pull request #%s"),
    "plymi": ("https://www.pythonlikeyoumeanit.com/%s", "%s"),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"
# html_logo = "../../brand/maite_logo_full_light_blue.png"

html_theme_options = {
    "collapse_navigation": True,
    "navigation_depth": 4,
    # "favicons": [
    #     {
    #         "rel": "icon",
    #         "sizes": "32x32",
    #         "href": "maite_favicon_32x32.png",
    #     },
    #     {
    #         "rel": "icon",
    #         "sizes": "64x64",
    #         "href": "maite_favicon_64x64.png",
    #     },
    # ],
    # "icon_links": [
    #     {
    #         "name": "GitHub",
    #         "url": "https://github.com/mit-ll-responsible-ai/maite",
    #         "icon": "fab fa-github-square",
    #     },
    # ],
}


def setup(app):
    app.add_js_file(
        "https://www.googletagmanager.com/gtag/js?id=UA-115029372-2",
        loading_method="async",
    )
    app.add_js_file("gtag.js")


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

myst_heading_anchors = 3



# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# Usage:
#
# python project_tooling/add_header.py
#
# Running this script will:
# 1. Remove OLD_HEADER from any targeted file that possesses that header
# 2. Add NEW_HEADER to the beginning of all targeted files
#
# See __main__ to see which directories/files are targeted by this script.
# Modify this section in the event that a new directory of .py files is
# added to the repo, which does not fall under any of the current directories.

import os
import os.path as path
from pathlib import Path
from typing import Iterable, Union

OLD_HEADER = """# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT\n\n"""

NEW_HEADER = """# Copyright 2023, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT\n\n"""

EXCLUDED = {"_version.py", "versioneer.py"}

OLD_HEADER = OLD_HEADER.splitlines(keepends=True)
NEW_HEADER = NEW_HEADER.splitlines(keepends=True)


def is_safe_dir(path: Path):
    path = path.resolve()
    if not Path.cwd() in path.parents:
        raise ValueError(
            f"Dangerous! You are running a script that can overwrite files in an unexpected directory: {path}"
        )


def get_src_files(dirname: Union[str, Path]) -> Iterable[Path]:
    dirname = Path(dirname)

    if dirname.is_file():
        if dirname.name.endswith(".py") and dirname.name not in EXCLUDED:
            yield dirname

    else:
        for cur, _, files in os.walk(dirname):
            cur = Path(cur)

            if any(p.startswith((".", "__")) for p in cur.parts):
                # exclude hidden/meta dirs
                continue

            for f in files:
                if f in EXCLUDED:
                    continue
                if f.endswith(".py"):
                    yield Path(path.join(cur, f))


def add_headers(files: Iterable[Path]):
    for file_ in files:
        is_safe_dir(file_)

        with file_.open("r") as f:
            contents = f.readlines()

        if contents[: len(OLD_HEADER)] == OLD_HEADER:
            contents = contents[len(OLD_HEADER) :]

        # if sum([len(line) for line in contents]) != 0:
        if contents[: len(NEW_HEADER)] != NEW_HEADER:
            contents = NEW_HEADER + contents

        with file_.open("w") as f:
            f.writelines(contents)


if __name__ == "__main__":
    add_headers(get_src_files("./src/maite/"))
    add_headers(get_src_files("./tests/"))



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from maite._internals.protocols.image_classification import (  # isort:skip
    InputType,
    TargetType,
    DatumMetadataType,
    InputBatchType,
    TargetBatchType,
    DatumMetadataBatchType,
    CollateFn,
    Augmentation,
    DataLoader,
    Dataset,
    Metric,
    Model,
)

__all__ = [
    "InputType",
    "TargetType",
    "DatumMetadataType",
    "InputBatchType",
    "TargetBatchType",
    "DatumMetadataBatchType",
    "CollateFn",
    "Augmentation",
    "DataLoader",
    "Dataset",
    "Metric",
    "Model",
]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

__all__ = ["MaiteException", "InternalError", "InvalidArgument"]


class MaiteException(Exception):
    # doc-ignore: NOQA
    """Base exception thrown by the MAITE."""


class InternalError(MaiteException, AssertionError):
    # doc-ignore: NOQA
    """An internal function was misconfigured or misused."""


class InvalidArgument(MaiteException, TypeError):
    # doc-ignore: NOQA
    """A MAITE interface was passed a bad value or type."""



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import TYPE_CHECKING

if not TYPE_CHECKING:
    try:
        from ._version import version as __version__
    except ImportError:
        __version__ = "unknown version"
else:  # pragma: no cover
    __version__: str



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# import component generics from generic.py and specialize them for image_classification
from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, Protocol, Sequence, Tuple

from typing_extensions import TypeAlias

from maite._internals.protocols import generic as gen
from maite.protocols import ArrayLike

# In below, the dimension names/meanings used are:
#
# N  - image instance
# H  - image height
# W  - image width
# C  - image channel
# Cl - classification label (one-hot for ground-truth label, or pseudo-probabilities for predictions)

InputType: TypeAlias = ArrayLike  # shape (C, H, W)
TargetType: TypeAlias = ArrayLike  # shape (Cl,)
DatumMetadataType: TypeAlias = Dict[str, Any]

InputBatchType: TypeAlias = Sequence[
    InputType
]  # sequence of N ArrayLikes of shape (C, H, W)
TargetBatchType: TypeAlias = Sequence[TargetType]  # sequence of N TargetType instances
DatumMetadataBatchType: TypeAlias = Sequence[DatumMetadataType]

Datum: TypeAlias = Tuple[InputType, TargetType, DatumMetadataType]
DatumBatch: TypeAlias = Tuple[InputBatchType, TargetBatchType, DatumMetadataBatchType]

CollateFn: TypeAlias = Callable[
    [Iterable[Datum]],
    DatumBatch,
]

# Initialize component classes based on generic and Input/Target/Metadata types


class Dataset(gen.Dataset[InputType, TargetType, DatumMetadataType], Protocol):
    """
    A dataset protocol for image classification ML subproblem providing datum-level
    data access.

    Implementers must provide index lookup (via `__getitem__(ind: int)` method) and
    support `len` (via `__len__()` method). Data elements looked up this way correspond
    to individual examples (as opposed to batches).

    Indexing into or iterating over the an image_classification dataset returns a
    `Tuple` of types `ArrayLike`, `ArrayLike`, and `Dict[str,Any]`.
    These correspond to the model input type, model target type, and datum-level
    metadata, respectively.

    Methods
    -------

    __getitem__(ind: int) -> Tuple[ArrayLike, ArrayLike, Dict[str, Any]]
        Provide map-style access to dataset elements. Returned tuple elements
        correspond to model input type, model target type, and datum-specific metadata type,
        respectively.

    __len__() -> int
        Return the number of data elements in the dataset.

    Examples
    --------

    We create a dummy set of data and use it to create a class that implements
    this lightweight dataset protocol:

    >>> import numpy as np
    >>> from typing import List, Dict, Any, Tuple
    >>> from maite.protocols import ArrayLike

    Assume we have 5 classes, 10 datapoints, and 10 target labels, and that we want
    to simply have an integer 'id' field in each datapoint's metadata:

    >>> N_CLASSES: int = 5
    >>> N_DATUM: int = 10
    >>> images: List[np.ndarray] = [np.random.rand(3, 32, 16) for _ in range(N_DATUM)]
    >>> targets: np.ndarray = np.eye(N_CLASSES)[np.random.choice(N_CLASSES, N_DATUM)]
    >>> metadata: List[Dict] = [{'id': i} for i in range(N_DATUM)]

    Constructing a compliant dataset just involves a simple wrapper that fetches
    individual datapoints, where a datapoint is a single image, target, metadata 3-tuple.

    >>> class ImageDataset:
    ...     def __init__(self,
    ...                  images: List[np.ndarray],
    ...                  targets: np.ndarray,
    ...                  metadata: List[Dict[str, Any]]):
    ...         self.images = images
    ...         self.targets = targets
    ...         self.metadata = metadata
    ...     def __len__(self) -> int:
    ...         return len(images)
    ...     def __getitem__(self, ind: int) -> Tuple[np.ndarray, np.ndarray, Dict[str, Any]]:
    ...         return self.images[ind], self.targets[ind], self.metadata[ind]

    We can instantiate this class and typehint it as an image_classification.Dataset.
    By using typehinting, we permit a static typechecker to verify protocol compliance.

    >>> from maite.protocols import image_classification as ic
    >>> dataset: ic.Dataset = ImageDataset(images, targets, metadata)

    Note that when writing a Dataset implementer, return types may be narrower than the
    return types promised by the protocol (np.ndarray is a subtype of ArrayLike), but
    the argument types must be at least as general as the argument types promised by the
    protocol.
    """

    ...


class DataLoader(
    gen.DataLoader[InputBatchType, TargetBatchType, DatumMetadataBatchType], Protocol
):
    """
    A dataloader protocol for the image classification ML subproblem providing
    batch-level data access.

    Implementers must provide an iterable object (returning an iterator via the
    `__iter__` method) that yields tuples containing batches of data. These tuples
    contain types `Sequence[ArrayLike]` (elements of shape `(C, H, W)`),
    `Sequence[ArrayLike]` (elements shape `(Cl, )`), and `Sequence[Dict[str, Any]]`,
    which correspond to model input batch, model target type batch, and a datum metadata batch.

    Note: Unlike Dataset, this protocol does not require indexing support, only iterating.

    Methods
    -------

    __iter__ -> Iterator[tuple[Sequence[ArrayLike], Sequence[ArrayLike], Sequence[Dict[str, Any]]]]
        Return an iterator over batches of data, where each batch contains a tuple of
        of model input batch (as `Sequence[ArrayLike]`), model target batch (as
        `Sequence[ArrayLike]`), and batched datum-level metadata
        (as `Sequence[Dict[str,Any]]`), respectively.

    """


class Model(gen.Model[InputBatchType, TargetBatchType], Protocol):
    """
    A model protocol for the image classification ML subproblem.

    Implementers must provide a `__call__` method that operates on a batch of model
    inputs (as `Sequence[ArrayLike]) and returns a batch of model targets (as
    `Sequence[ArrayLike]`)

    Methods
    -------

    __call__(input_batch: Sequence[ArrayLike]) -> Sequence[ArrayLike]
        Make a model prediction for inputs in input batch. Input batch is expected to
        be `Sequence[ArrayLike]` with each element of shape `(C, H, W)`.
    """


class Metric(gen.Metric[TargetBatchType], Protocol):
    """
    A metric protocol for the image classification ML subproblem.

    A metric in this sense is expected to measure the level of agreement between model
    predictions and ground-truth labels.

    Methods
    -------

    update(preds: Sequence[ArrayLike], targets: Sequence[ArrayLike]) -> None
        Add predictions and targets to metric's cache for later calculation. Both
        preds and targets are expected to be sequences with elements of shape `(Cl,)`.

    compute() -> Dict[str, Any]
        Compute metric value(s) for currently cached predictions and targets, returned as
        a dictionary.

    reset() -> None
        Clear contents of current metric's cache of predictions and targets.
    """

    ...


class Augmentation(
    gen.Augmentation[
        InputBatchType,
        TargetBatchType,
        DatumMetadataBatchType,
        InputBatchType,
        TargetBatchType,
        DatumMetadataBatchType,
    ],
    Protocol,
):
    """
    An augmentation protocol for the image classification subproblem.

    An augmentation is expected to take a batch of data and return a modified version of
    that batch. Implementers must provide a single method that takes and returns a
    labeled data batch, where a labeled data batch is represented by a tuple of types
    `Sequence[ArrayLike]` (with elements of shape `(C, H, W)`), `Sequence[ArrayLike]`
    (with elements of shape `(Cl, )`), and `Sequence[Dict[str,Any]]`. These correspond
    to the model input batch type, model target batch type, and datum-level metadata
    batch type, respectively.

    Methods
    -------

    __call__(datum: Tuple[Sequence[ArrayLike], Sequence[ArrayLike], Sequence[dict[str, Any]]]) ->\
          Tuple[Sequence[ArrayLike], Sequence[ArrayLike], Sequence[dict[str, Any]]])
        Return a modified version of original data batch. A data batch is represented
        by a tuple of model input batch (as `Sequence[ArrayLike]` with elements of shape
        `(C, H, W)`), model target batch (as an `Sequence[ArrayLike]` of shape `(N, Cl)`),
        and batch metadata (as `Sequence[Dict[str, Any]]`), respectively.

    Examples
    --------

    We can write an implementer of the augmentation class as either a function or a class.
    The only requirement is that the object provide a __call__ method that takes objects
    at least as general as the types promised in the protocol signature and return types
    at least as specific.

    >>> import copy
    >>> import numpy as np
    >>> from typing import Dict, Any, Tuple, Sequence
    >>> from maite.protocols import ArrayLike
    >>>
    >>> class ImageAugmentation:
    ...     def __call__(
    ...         self,
    ...         data_batch: Tuple[Sequence[ArrayLike], Sequence[ArrayLike], Sequence[Dict[str, Any]]]
    ...     ) -> Tuple[Sequence[np.ndarray], Sequence[np.ndarray], Sequence[Dict[str, Any]]]:
    ...         inputs, targets, mds = data_batch
    ...         # We copy data passed into the constructor to avoid mutating original inputs
    ...         # By using np.ndarray constructor, the static type-checker will let us treat
    ...         # generic ArrayLike as a more narrow return type
    ...         inputs_aug = [copy.copy(np.array(input)) for input in inputs]
    ...         targets_aug = [copy.copy(np.array(target)) for target in targets]
    ...         mds_aug = copy.deepcopy(mds)  # deepcopy in case of nested structure
    ...         # Modify inputs_aug, targets_aug, or mds_aug as needed
    ...         # In this example, we just add a new metadata field
    ...         for i, md in enumerate(mds_aug):
    ...             md['new_key'] = i
    ...         return inputs_aug, targets_aug, mds_aug

    We can typehint an instance of the above class as an Augmentation in the
    image_classification domain:

    >>> from maite.protocols import image_classification as ic
    >>> im_aug: ic.Augmentation = ImageAugmentation()

    """



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import random

import numpy as np
import torch
from hypothesis import given


@given(...)
def test_python_random_controlled_by_hypothesis(dummy: None):
    # hypothesis sets initial seed to 0 within scope of test
    pre_seed = random.random()
    random.seed(0)
    assert pre_seed == random.random()


@given(...)
def test_torch_random_controlled_by_hypothesis(dummy: None):
    # hypothesis sets initial seed to 0 within scope of test
    pre_seed = torch.rand((1,)).item()
    torch.manual_seed(0)
    assert pre_seed == torch.rand((1,)).item()


@given(...)
def test_numpy_random_controlled_by_hypothesis(dummy: None):
    # hypothesis sets initial seed to 0 within scope of test
    pre_seed = np.random.random((1,)).item()
    np.random.seed(0)
    assert pre_seed == np.random.random((1,)).item()



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import json
import subprocess
from copy import deepcopy
from functools import _CacheInfo as CacheInfo, lru_cache
from importlib import import_module
from pathlib import Path
from typing import Any, Collection, FrozenSet, Generator, List, Literal, Mapping, Union

from typing_extensions import NotRequired, TypedDict

from maite._internals.testing.pyright import PYRIGHT_PATH, Diagnostic, PyrightOutput
from maite.errors import InvalidArgument
from maite.utils.validation import check_one_of, check_type

from ..utils import is_typed_dict

Category = Literal[
    "class",
    "constant",
    "function",
    "method",
    "module",
    "symbol",
    "type alias",
    "variable",
]

CATEGORIES: FrozenSet[Category] = frozenset(Category.__args__)  # type: ignore


class Symbol(TypedDict):
    category: Category
    name: str
    referenceCount: int
    isExported: bool
    isTypeKnown: bool
    isTypeAmbiguous: bool
    diagnostics: List[Diagnostic]


class SymbolCounts(TypedDict):
    withKnownType: int
    withAmbiguousType: int
    withUnknownType: int


class CompletenessSection(TypedDict):
    packageName: str
    packageRootDirectory: Path
    moduleName: str
    moduleRootDirectory: Path
    ignoreUnknownTypesFromImports: bool
    pyTypedPath: NotRequired[Path]
    exportedSymbolCounts: SymbolCounts
    otherSymbolCounts: SymbolCounts
    missingFunctionDocStringCount: int
    missingClassDocStringCount: int
    missingDefaultParamCount: int
    completenessScore: float
    modules: List[Any]
    symbols: List[Symbol]


class ModuleScanResults(PyrightOutput):
    """The schema for the JSON output of a type completeness scan.

    The output type of `maite.testing.project.ModuleScan.__call__"""

    # doc-ignore: NOQA
    typeCompleteness: CompletenessSection


def _pyright_type_completeness(
    module_name: str,
    *,
    path_to_pyright: Union[Path, None],
) -> ModuleScanResults:
    module_name = module_name.replace("-", "_")

    if path_to_pyright is None:  # pragma: no cover
        raise ModuleNotFoundError(
            "`pyright` was not found. It may need to be installed."
        )
    if not path_to_pyright.is_file():
        raise FileNotFoundError(
            f"`path_to_pyright` – {path_to_pyright} – doesn't exist."
        )

    proc = subprocess.run(
        [
            str(PYRIGHT_PATH),
            "--ignoreexternal",
            "--outputjson",
            "--verifytypes",
            module_name,
        ],
        cwd=Path.cwd(),
        encoding="utf-8",
        text=True,
        capture_output=True,
    )
    try:
        out = json.loads(proc.stdout)
    except Exception as e:  # pragma: no cover
        print(proc.stdout)
        raise e

    scan_section = out["typeCompleteness"]
    for k in ["packageRootDirectory", "moduleRootDirectory", "pyTypedPath"]:
        if k in scan_section:
            scan_section[k] = Path(scan_section[k])

    return out


class ModuleScan:
    """
    Uses pyright's type completeness scan to summarize a module's contents.

    By default, `ModuleScan`'s __call__ is cached to reduce overhead for getting
    scan results for a module multiple times. Each `ModuleScan` instance has an
    independent cache.

    Methods
    -------
    __call__(module_name, *, path_to_pyright=PYRIGHT_PATH, cached=True)
        Perform a scan on the specified module or submodule.

    cache_info()
        Return a CacheInfo object with statistics about the cache of this instance.

    Examples
    --------
    >>> from maite.testing.project import ModuleScan
    >>> scanner = ModuleScan()
    >>> results = scanner("maite")
    >>> results["summary"]
    {'filesAnalyzed': ..., 'errorCount': ..., 'warningCount': ..., 'informationCount': ..., 'timeInSec': ...}
    >>> results["typeCompleteness"]["packageName"]
    'maite'
    >>> results["typeCompleteness"]["symbols"]
    [{'category': 'class',
      'name': 'maite.errors.MaiteException',
      'referenceCount': 3,
      'isExported': True,
      'isTypeKnown': True,
      'isTypeAmbiguous': False,
      'diagnostics': []},
     {'category': 'class',
      'name': 'maite.errors.InternalError',
      'referenceCount': 1,
      'isExported': True,
      'isTypeKnown': True,
      'isTypeAmbiguous': False,
      'diagnostics': []},
    ...
    """

    def __init__(self) -> None:
        self._cached_scan = lru_cache(maxsize=256, typed=False)(
            _pyright_type_completeness
        )

    def __call__(
        self,
        module_name: str,
        *,
        path_to_pyright: Union[Path, None] = PYRIGHT_PATH,
        cached: bool = True,
    ) -> ModuleScanResults:
        # doc-ignore: EX01
        """
        Perform a scan on the specified module or submodule.

        Parameters
        ----------
        module_name : str
            The name of a module to scan. This must the name of a module that is
            installed in the current Python environment.

        path_to_pyright : Path, optional, keyword-only
            Path to the pyright executable (see installation instructions).
            Defaults to `shutil.where('pyright')` if the executable can be found.

        cached : bool, optional (default=True)
            If `True`, then the result of the scan will be mediated by an LRU cache.
            I.e. subsequent identical calls will return a cached result.

        Returns
        -------
        ModuleScanResults
            A dictionary containing::

                version: str
                time: str
                generalDiagnostics: List[Diagnostic]  (empty)
                summary: Summary
                    filesAnalyzed: int
                    errorCount: int
                    warningCount: int
                    informationCount: int
                    timeInSec: float
                typeCompleteness: CompletenessSection
                    packageName: str
                    packageRootDirectory: Path
                    moduleName: str
                    moduleRootDirectory: Path
                    ignoreUnknownTypesFromImports: bool
                    pyTypedPath: NotRequired[Path]
                    exportedSymbolCounts: dict
                    otherSymbolCounts: dict
                    missingFunctionDocStringCount: int
                    missingClassDocStringCount: int
                    missingDefaultParamCount: int
                    completenessScore: float
                    modules: list
                    symbols: List[Symbol]

        Raises
        ------
        ModuleNotFoundError
            `module` was not found.
        """
        scan = self._cached_scan if cached else _pyright_type_completeness
        out = scan(module_name, path_to_pyright=path_to_pyright)
        _summary = out["summary"]
        # TODO: If a project does not have a py.typed file then no files will be analyzed.
        #       This does *not* mean that the module was not found. Instead, we should
        #       raise an error that explicitly states that the project was found be requires
        #       a py.typed file in order to be analyzed
        if _summary["errorCount"] > 0 and _summary["filesAnalyzed"] == 0:
            raise ModuleNotFoundError(
                f"No files were found to analyze in association "
                f"with module `{module_name}`. The module may not be installed or "
                f"there may be a typo in the name."
            )

        return deepcopy(out) if cached else out

    def clear_cache(self) -> None:
        # doc-ignore: GL08
        self._cached_scan.cache_clear()

    def cache_info(self) -> CacheInfo:
        """
        Access the scanner instance's cache stats.

        Returns
        -------
        CacheInfo
            A named tuple with fields:
            - hits
            - misses
            - maxsize
            - currsize
        """
        # doc-ignore: NOQA
        return self._cached_scan.cache_info()


def _is_dunder(name: str) -> bool:
    *_, name = name.split(".")
    out = (
        name.startswith("__")
        and name.endswith("__")
        and not name.startswith("___")
        and not name.endswith("___")
    )
    return out


def get_public_symbols(
    scan: ModuleScanResults, submodule: str = "", include_dunder_names: bool = False
) -> List[Symbol]:
    """
    Return all public symbols (functions, classes, etc.) from a module's API.

    This function expects the results of a scan performed by
    `maite.testing.project.ModuleScan`, which requires that `pyright` is
    installed.

    Parameters
    ----------
    scan : ModuleScanResults
        The result of a scan performed by `ModuleScan.__call__`.

    submodule : str, optional
        If specified, only symbols from the specified submodule are included.

    include_dunder_names : bool, default=False
        If `True`, then symbols like the `__init__` method of public classes will
        be included among the public symbols.

    Returns
    -------
    List[Symbol]
        Each symbol is a dict containing the following key-value pairs::

            category: Literal["class", "constant", "function", "method",
                                "module", "symbol", "type alias", "variable"]
            name: str
            referenceCount: int
            isExported: bool
            isTypeKnown: bool
            isTypeAmbiguous: bool
            diagnostics: List[Diagnostic]

    Examples
    --------
    Basic usage.

    >>> from maite.testing.project import get_public_symbols, ModuleScan
    >>> scanner = ModuleScan()
    >>> results = scanner("maite")
    >>> get_public_symbols(results)
    [ {'category': 'class',
    'name': 'maite.testing.project.ModuleScan',
    'referenceCount': 1,
    'isExported': True,
    'isTypeKnown': False,
    'isTypeAmbiguous': False,
    'diagnostics': []},
    {'category': 'class',
    'name': 'maite.testing.project.ModuleScanResults',
    'referenceCount': 1,
    'isExported': True,
    'isTypeKnown': False,
    'isTypeAmbiguous': False,
    'diagnostics': []},
    {'category': 'class',
    'name': 'maite.testing.project.Symbol',
    'referenceCount': 1,
    'isExported': True,
    'isTypeKnown': True,
    'isTypeAmbiguous': False,
    'diagnostics': []},
    ...
    ]

    Accessing symbols from the `docs` submodule.

    >>> get_public_symbols(results, submodule="maite.testing.docs")
    [{'category': 'type alias', 'name': 'maite.testing.docs.NumpyDocErrorCode', 'referenceCount': 1, 'isExported': True, 'isTypeKnown': True, 'isTypeAmbiguous': False, 'diagnostics': []}, {'category': 'class', 'name': 'maite.testing.docs.NumPyDocResults', 'referenceCount': 1, 'isExported': True, 'isTypeKnown': True, 'isTypeAmbiguous': False, 'diagnostics'...
    """
    check_type("scan", scan, Mapping)
    check_type("submodule", submodule, str)
    check_type("include_dunder_names", include_dunder_names, bool)

    out = (x for x in scan["typeCompleteness"]["symbols"] if x["isExported"])
    if not include_dunder_names:
        # It is overly restrictive to only check for methods.
        # E.g. a dataclass' `__new__` method is actually
        # categorized as a function
        out = (x for x in out if not _is_dunder(x["name"]))
    if submodule:
        if any(not x.isidentifier() for x in submodule.split(".")):
            raise InvalidArgument(f"{submodule} is not a valid module name.")

        _out = [x for x in out if x["name"].startswith(submodule)]

        return _out
    return list(out)


def import_public_symbols(
    scan: ModuleScanResults,
    submodule: str = "",
    categories: Collection[Category] = frozenset(["function", "class"]),
    skip_module_not_found: Union[bool, Literal["pytest-skip"]] = True,
) -> Generator[Any, None, None]:
    """
    Import and yield all public symbols (functions, classes, etc.) from a module's API.

    This function expects the results of a scan performed by
    `maite.testing.project.ModuleScan`, which requires that `pyright` is
    installed.

    Parameters
    ----------
    scan : ModuleScanResults
        The result of a scan performed by `ModuleScan.__call__`.

    submodule : str, optional
        If specified, only symbols from the specified submodule are included.

    categories : Collection[Category], default=('function', 'class')
        The symbol categories to include.

        Valid categories are:
           - class
           - constant
           - function
           - method
           - module
           - symbol
           - type alias
           - variable

    skip_module_not_found : bool | Literal["pytest-skip"], default=True
        If `True`, symbols that cannot be imported due to a `ModuleNotFound` error
        are skipped (e.g., due to missing optional dependencies). If `False`, the
        `ModuleNotFound` error is raised.

        If `'pytest-skip'` is specified a pytest param of the symbol name – marked
        by pytest to skip – is yielded when the import raises a `ModuleNotFound` error.
        This is useful when `import_public_symbols` is used to populate a parameterized
        test, so that skipped symbols are documented.

    Yields
    ------
    Any
        The imported symbol.

    Examples
    --------
    Basic usage.

    >>> from maite.testing.project import import_public_symbols, ModuleScan
    >>> scanner = ModuleScan()
    >>> results = scanner("pyright")
    >>> list(import_public_symbols(results))[:2]
    [<function pyright.cli.entrypoint() -> NoReturn>,
    <function pyright.cli.main(args: List[str], **kwargs: Any) -> int>]
    """

    if isinstance(skip_module_not_found, bool):
        marker = None
    elif skip_module_not_found == "pytest-skip":
        # pytest should be kept optional.
        from pytest import mark, param

        marker = mark.skip(reason="Module not found.")
    else:
        check_one_of(
            "skip_module_not_found", skip_module_not_found, [True, False, "pytest-skip"]
        )
        marker = None  # pragma: no cover
        raise Exception("unreachable")  # pragma: no cover

    for cat in categories:
        check_one_of("categories", cat, CATEGORIES)

    categories = set(categories)

    symbols = filter(
        lambda symbol: symbol["category"] in categories,
        sorted(
            get_public_symbols(scan, submodule=submodule),
            key=lambda symbol: symbol["name"],
        ),
    )

    err = (ModuleNotFoundError, AttributeError) if skip_module_not_found else ()
    cached_typeddict_names = set()

    for symbol in symbols:
        module_path, name = symbol["name"].rsplit(".", maxsplit=1)

        # TODO: probably need a more sophisticated method for importing
        # e.g. https://github.com/facebookresearch/hydra/blob/9ce67207488965431c69b2e2b8e1a2baa0ada4b8/hydra/_internal/utils.py#L614
        #
        # For example, if 'method' is included in `categories` then we can cash each
        # class-object that we import and then getattr on it when we encounter a
        # method/variable on that class. Because the symbols are sorted alphabetically
        # we are always guaranteed to encounter a class object before its members.

        if (
            symbol["category"] == "function"
            and module_path in cached_typeddict_names
            and "method" not in categories
        ):
            continue

        try:
            module = import_module(module_path)
        except err:
            if marker is not None:
                yield param(symbol["name"], marks=marker)  # type: ignore

            else:
                continue
        else:
            try:
                obj = getattr(module, name)
                if symbol["category"] == "class" and is_typed_dict(obj):
                    cached_typeddict_names.add(symbol["name"])

                yield obj
            except err:  # pragma: no cover
                # it is possible that a symbol is unreachable at runtime
                if marker is not None:
                    yield param(symbol["name"], marks=marker)  # type: ignore
                else:
                    continue



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# This module is designed to intentionally include dependencies (i.e. mygrad)
# that have not been installed by default. maite should handle this
# gracefully.

from mygrad import tensor

__all__ = ["func_needs_mygrad"]


def func_needs_mygrad() -> None:
    tensor(1)
    return None



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from maite._internals.import_utils import is_hypothesis_available

if is_hypothesis_available():
    from maite._internals.testing.hypothesis import image_data
else:
    raise ImportError(
        "maite.testing.hypothesis requires that hypothesis be installed as a dependency."
    )  # pragma: no cover

__all__ = ["image_data"]



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any

import hypothesis.strategies as st
import numpy as np
from hypothesis.strategies import DrawFn
from numpy.typing import NDArray


# Define a strategy for generating random image data
def image_data(
    draw: DrawFn, channels: int = 3, max_width: int = 10, max_height: int = 10
) -> NDArray[Any]:
    """
    Create a strategy for generating random image data.

    Parameters
    ----------
    draw : Callable
        The Hypothesis draw function.
    channels : int, optional
        The number of channels in the image, by default 3.
    max_width : int, optional
        The maximum width of the image, by default 10.
    max_height : int, optional
        The maximum height of the image, by default 10.

    Returns
    -------
    np.ndarray
        A random image array.

    Examples
    --------
    >>> import hypothesis.strategies as st
    >>> from maite.testing.hypothesis import image_data
    >>> st.composite(image_data)().example()
    array([[[  0,  10,  20]]])
    """
    # Generate a random width and height for the image
    width = draw(st.integers(min_value=1, max_value=max_width))
    height = draw(st.integers(min_value=1, max_value=max_height))

    # Generate random pixel values for the image
    pixels = draw(
        st.lists(
            elements=st.integers(min_value=0, max_value=255),
            min_size=width * height * channels,
            max_size=width * height * channels,
        )
    )

    # Reshape the pixel values into an image array
    image_array = np.array(pixels).reshape((height, width, channels))

    return image_array



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Dict, List, Type, TypeVar

from typing_extensions import TypedDict, TypeGuard

K = TypeVar("K")
T = TypeVar("T")
Td = TypeVar("Td", bound=TypedDict)


def is_list_of_type(d: Any, guard: Type[T] = Any) -> TypeGuard[List[T]]:
    """
    Check if object is a list of dictionaries.

    Parameters
    ----------
    d : Any
        The object to check.
    guard : Type[T]
        The type guard of the dictionaries. Defaults to dict.

    Returns
    -------
    TypeGuard[List[T]]
        True if object is a list of dictionaries.

    Examples
    --------
    >>> is_list_dict([{"a": 1}, {"b": 2}])
    True
    """
    return isinstance(d, (list, tuple)) and isinstance(d[0], guard)


def is_list_dict(d: Any, guard: Type[T] = Dict[Any, Any]) -> TypeGuard[List[T]]:
    """
    Check if object is a list of dictionaries.

    Parameters
    ----------
    d : Any
        The object to check.
    guard : Type[T]
        The type guard of the dictionaries. Defaults to dict.

    Returns
    -------
    TypeGuard[List[T]]
        True if object is a list of dictionaries.

    Examples
    --------
    >>> is_list_dict([{"a": 1}, {"b": 2}])
    True
    """
    return isinstance(d, (list, tuple)) and isinstance(d[0], dict)


def is_typed_dict(object: Any, target: Type[Td]) -> TypeGuard[Td]:
    """
    Check if object is a typed dictionary.

    Parameters
    ----------
    object : Any
        The object to check.

    target : Type[T]
        The type of the dictionary.

    Returns
    -------
    TypeGuard[T]
        True if object is a typed dictionary.

    Examples
    --------
    >>> from typing_extensions import TypedDict
    >>> class Foo(TypedDict):
    ...     a: int
    >>> is_typed_dict({"a": 1}, Foo)
    True
    """
    if not isinstance(object, dict):
        return False

    k_obj = set(object.keys())
    ks = set(target.__annotations__.keys())

    if hasattr(target, "__total__") and target.__total__:
        return all(k in k_obj for k in ks)
    else:
        return any(k in k_obj for k in ks)



# Copyright 2024, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# pyright: strict

from enum import Enum, EnumMeta
from itertools import chain
from typing import (
    Any,
    Callable,
    Collection,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
    runtime_checkable,
)

from maite.errors import InvalidArgument

T = TypeVar("T")
N = TypeVar("N", int, float)
C = TypeVar("C", bound="Comparable")


def _tick(x: str):
    return "`" + x + "`"


def _safe_name(x: Any, ticked: bool = True) -> str:
    out = getattr(x, "__name__", str(x))
    return _tick(out) if ticked else out


@runtime_checkable
class Comparable(Protocol):
    def __eq__(self, __x: Any) -> bool:
        ...

    def __lt__(self: C, __x: C) -> bool:
        ...

    def __gt__(self: C, __x: C) -> bool:
        ...

    def __le__(self: C, __x: C) -> bool:
        ...

    def __ge__(self: C, __x: C) -> bool:
        ...


class Unsatisfiable(AssertionError):
    ...


def check_type(
    name: str, arg: T, type_: Union[type, Tuple[type, ...]], *, optional: bool = False
) -> T:
    """
    Check that an argument is an instance of one or more types.

    Parameters
    ----------
    name : str
        The argument's name.

    arg : T (Any)
        The argument.

    type_ : type | tuple[type, ...]
        The type that `arg` should belong to. If multiple types are provided
        then checks that `arg` is an instance of at least one of them.

    optional : bool, keyword, optional (default=False)
        If ``True``, then ``arg`` can be None.

    Returns
    -------
    T
        The input value, `arg`, unchanged.

    Raises
    ------
    InvalidArgument
        `arg` is not of the expected type.

    Examples
    --------
    >>> from maite.utils.validation import check_type
    >>> check_type('apple', 1, int)
    1

    >>> try:
    ...     check_type('apple', 1, bool)
    ... except:
    ...     print("maite.errors.InvalidArgument: Expected `apple` to be of type `bool`. Got `1` (type: `int`).")
    maite.errors.InvalidArgument: Expected `apple` to be of type `bool`. Got `1` (type: `int`).

    >>> check_type('apple', 1, (int, bool))
    1

    >>> print(check_type('apple', None, (int, bool), optional=True))
    None
    """
    if optional and arg is None:
        return arg

    if not isinstance(arg, type_):
        if isinstance(type_, tuple):
            clause = f"of types: {', '.join(_safe_name(t) for t in type_)}"
        else:
            clause = f"of type {_safe_name(type_)}"

        raise InvalidArgument(
            f"Expected `{name}` to be {'`None` or ' if optional else ''}{clause}. Got "
            f"`{arg}` (type: {_safe_name(type(arg))})."
        )
    return arg


@overload
def check_domain(
    name: str,
    arg: C,
    *,
    lower: Comparable,
    upper: Optional[Comparable] = None,
    incl_low: bool = ...,
    incl_up: bool = ...,
    lower_name: str = ...,
    upper_name: str = ...,
) -> C:
    ...


@overload
def check_domain(
    name: str,
    arg: C,
    *,
    lower: Optional[Comparable] = None,
    upper: Comparable,
    incl_low: bool = ...,
    incl_up: bool = ...,
    lower_name: str = ...,
    upper_name: str = ...,
) -> C:
    ...


def check_domain(
    name: str,
    arg: C,
    *,
    lower: Optional[Comparable] = None,
    upper: Optional[Comparable] = None,
    incl_low: bool = True,
    incl_up: bool = True,
    lower_name: str = "",
    upper_name: str = "",
) -> C:
    """
    Check that an argument falls within `[lower <=] arg [<= upper]`.

    Parameters
    ----------
    name : str
        The argument's name.

    arg : Comparable
        The value to be checked.

    lower : Optional[Comparable]
        The lower bound of the domain. This bound is not checked
        if unspecified.

    upper : Optional[Comparable]
        The upper bound of the domain. This bound is not checked
        if unspecified.

    incl_low : bool, optional (default=True)
        If `True`, the lower bound is inclusive.

    incl_up : bool, optional (default=True)
        If `True`, the upper bound is inclusive.

    lower_name : str = ""
        If specified, includes the name of the lower bound in the
        error message.

    upper_name : str = ""
        If specified, includes the name of the upper bound in the
        error message.

    Returns
    -------
    Comparable
        The input value, `arg`, unchanged.

    Raises
    ------
    InvalidArgument
        `arg` does not satisfy the inequality.

    Unsatisfiable
        An internal assertion error when the provided domain
        bounds cannot be satisfied.

    Examples
    --------
    >>> from maite.utils.validation import check_domain
    >>> try:
    ...     check_domain("x", 1, lower=20)
    ... except:
    ...     print("maite.errors.InvalidArgument: `x` must satisfy `20 <= x`.  Got: `1`.")
    maite.errors.InvalidArgument: `x` must satisfy `20 <= x`.  Got: `1`.

    >>> try:
    ...     check_domain("x", 1, lower=1, incl_low=False)
    ... except:
    ...     print("maite.errors.InvalidArgument: `x` must satisfy `1 < x`.  Got: `1`.")
    maite.errors.InvalidArgument: `x` must satisfy `1 < x`.  Got: `1`.

    >>> check_domain("x", 1, lower=1, incl_low=True) # ok
    1

    >>> check_domain("x", 0.0, lower=-10, upper=10)  # ok
    0.0
    """
    # check internal params
    check_type("name", name, str)
    check_type("incl_low", incl_low, bool)
    check_type("incl_up", incl_up, bool)

    if lower is not None and upper is not None:
        if incl_up and incl_low:
            if not (lower <= upper):
                raise Unsatisfiable(f"{lower} <= {upper}")
        elif not lower < upper:
            raise Unsatisfiable(f"{lower} < {upper}")
    elif lower is None and upper is None:
        raise Unsatisfiable("Neither `lower` nor `upper` were specified.")

    min_satisfied = (
        (lower <= arg if incl_low else lower < arg) if lower is not None else True
    )
    max_satisfied = (
        (arg <= upper if incl_up else arg < upper) if upper is not None else True
    )

    if not min_satisfied or not max_satisfied:
        lsymb = "<=" if incl_low else "<"
        rsymb = "<=" if incl_up else "<"

        err_msg = f"`{name}` must satisfy `"

        if lower is not None:
            if lower_name:
                lower = f"{lower_name}={lower}"
            err_msg += f"{lower} {lsymb} "

        err_msg += f"{name}"

        if upper is not None:
            if upper_name:
                upper = f"{upper_name}={upper}"
            err_msg += f" {rsymb} {upper}"

        err_msg += f"`.  Got: `{arg}`."

        raise InvalidArgument(err_msg)
    return arg


class SupportsEq(Protocol):
    def __eq__(self, __o: object) -> bool:
        ...


def check_one_of(
    name: str,
    arg: T,
    collection: Union[Collection[Any], Type[Enum]],
    *vals: SupportsEq,
    requires_identity: bool = False,
) -> T:
    """
    Check that `arg` is a member of `collection` or of `vals`.

    Parameters
    ----------
    name : str
        The argument's name.

    arg : T (Any)
        The argument.

    collection : Collection | Type[Enum]
        Any collection (i.e., supports `__contains__` and `__iter__`) or enum type.

    *vals : Any
        Additional values to check `arg` against.

    requires_identity : bool, optional (default=False)
        If `True`, all (non hash-based) membership checks are performed element-wise
        using `x is e` rather  than the default `x is e or x == e`.

        This can be helpful for ensuring strict identity, e.g., preventing `1` from
        matching against `True`.

    Returns
    -------
    T
        The input value, `arg`, unchanged.

    Raises
    ------
    InvalidArgument
        `arg` is not of a member of `collections` nor `vals`.

    Unsatisfiable
        An internal assertion error when the provided collection is empty.

    Examples
    --------
    >>> from maite.utils.validation import check_one_of
    >>> try:
    ...     check_one_of("foo", None, [1, 2])
    ... except:
    ...     print("maite.errors.InvalidArgument: Expected `foo` to be one of: 1, 2. Got `None`.")
    maite.errors.InvalidArgument: Expected `foo` to be one of: 1, 2. Got `None`.

    Including `None` as an acceptable value

    >>> print(check_one_of("foo", None, [1, 2], None))
    None

    Enforcing strict identity using `requires_identity`:

    >>> check_one_of("foo", 1, [True])  # `1` == `True`
    1

    >>> try:
    ...     check_one_of("foo", 1, [True], requires_identity=True)
    ... except:
    ...     print("maite.errors.InvalidArgument: Expected `foo` to be: True. Got `1`.")
    maite.errors.InvalidArgument: Expected `foo` to be: True. Got `1`.

    Support for enums:

    >>> from enum import Enum
    >>> class Pet(Enum):
    ...     cat = 1
    ...     dog = 2

    >>> try:
    ...     check_one_of("bar", 88, Pet)
    ... except:
    ...     print("maite.errors.InvalidArgument: Expected `bar` to be one of: Pet.cat, Pet.dog. Got `88`.")
    maite.errors.InvalidArgument: Expected `bar` to be one of: Pet.cat, Pet.dog. Got `88`.
    >>> check_one_of("bar", Pet.cat, Pet)
    <Pet.cat: 1>
    """

    if isinstance(collection, EnumMeta):
        if isinstance(arg, collection):
            return arg
    elif requires_identity:
        collection = cast(Collection[Any], collection)
        if any(arg is x for x in chain(collection, vals)):
            return arg
    elif arg in collection or arg in vals:
        return arg

    values = sorted(set(str(x) for x in chain(collection, vals)))
    if not values:
        raise Unsatisfiable("`collections` and `args` are both empty.")

    raise InvalidArgument(
        f"Expected `{name}` to be{' one of' if len(values) > 1 else ''}: "
        f"{', '.join(values)}. Got `{arg}`."
    )


def chain_validators(*validators: Callable[[str, Any], Any]) -> Callable[[str, T], T]:
    """
    Enable validators to be chained together.

    Validators are functions like `(name: str, arg: T, [...]) -> T`. This is meant to be
    used with partial'd validators, where only the `name` and `arg` fields need be
    populated.

    Parameters
    ----------
    *validators : Callable[[str, T], T]
        Accepts `name` and `arg`, and returns `arg` if it is a valid input, otherwise
        should raise `InvalidArgument`.

    Returns
    -------
    Callable[[str, T], T]
        A function of signature `chain(name: str, arg: T) -> T` that calls each
        validator as `val(name, arg)` in order from low-index to high-index.

    Examples
    --------
    >>> from maite.utils.validation import check_type, check_domain, chain_validators
    >>> from functools import partial
    >>> is_int = partial(check_type, type_=int)
    >>> is_pos = partial(check_domain, lower=0)
    >>> check_pos_int = chain_validators(is_int, is_pos)
    >>> check_pos_int("foo", 10)
    10
    >>> try:
    ...     check_pos_int("foo", ["a"])
    ... except:
    ...     print("maite.errors.InvalidArgument: Expected `foo` to be of type `int`. Got `['a']` (type: `list`).")
    maite.errors.InvalidArgument: Expected `foo` to be of type `int`. Got `['a']` (type: `list`).
    >>> try:
    ...     check_pos_int("foo", -1)
    ... except:
    ...     print("maite.errors.InvalidArgument: `foo` must satisfy `0 <= foo`.  Got: `-1`.")
    maite.errors.InvalidArgument: `foo` must satisfy `0 <= foo`.  Got: `-1`.
    """

    def chain(name: str, arg: T) -> T:
        for v in validators:
            v(name, arg)
        return arg

    return chain
